import asyncio
import json
import logging
import sys
from typing import List, Optional

import aiohttp
import litellm
import numpy as np
from importlib_resources import files
from jinja2 import Environment, FileSystemLoader
from pydantic_core import ValidationError
from tqdm.asyncio import tqdm

from grouse.dtos import (
    AnswerRelevancy,
    AnswerRelevancyPair,
    Completeness,
    CompletenessPair,
    EvaluationSample,
    EvaluationsAndReport,
    Failed,
    Faithfulness,
    FaithfulnessPair,
    GroundedQAEvaluation,
    GroundedQAEvaluationReport,
    Score,
    ScorePair,
    Usefulness,
    UsefulnessPair,
)
from grouse.utils import get_positive_acceptance_negative_rejection

STRUCTURED_OUTPUTS_SUPPORTING_MODELS = [
    "gpt-4o-mini-2024-07-18",
    "gpt-4o-2024-08-06",
]


class GroundedQAEvaluator:
    def __init__(
        self,
        model_name: str = "gpt-4",
        prompts_path: Optional[str] = None,
        cache_path: Optional[str] = None,
    ):
        self.model_name = model_name
        if prompts_path is None:
            self.environment = Environment(
                loader=FileSystemLoader(files("grouse").joinpath("gpt4_prompts"))
            )
        else:
            self.environment = Environment(loader=FileSystemLoader(prompts_path))

        self.logger = logging.getLogger("LLM Call Tracker")
        self.logger.setLevel(logging.INFO)
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.INFO)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

        litellm.enable_cache("disk", cache_path)

        self.cost = 0

    async def call_llm(self, prompt: str, pair_model: ScorePair) -> Score | Failed:
        try:
            kwargs = {"temperature": 0.01, "max_tokens": 2048}
            if self.model_name in STRUCTURED_OUTPUTS_SUPPORTING_MODELS:
                response = await litellm.acompletion(
                    model=self.model_name,
                    messages=[{"role": "user", "content": prompt}],
                    response_format=pair_model,
                    **kwargs,
                )
            elif "-turbo" in self.model_name or "4o" in self.model_name:
                response = await litellm.acompletion(
                    model=self.model_name,
                    messages=[{"role": "user", "content": prompt}],
                    response_format={"type": "json_object"},
                    **kwargs,
                )
            else:
                response = await litellm.acompletion(
                    model=self.model_name,
                    messages=[{"role": "user", "content": prompt}],
                    **kwargs,
                )
            loaded_response = json.loads(response.choices[0].message.content)
            if isinstance(loaded_response, dict):
                pair = pair_model(**loaded_response)
                self.cost += litellm.completion_cost(response)
                return pair.answer_2
            else:
                raise ValueError("Response is not a dictionary")

        except (ValidationError, json.decoder.JSONDecodeError, ValueError) as val_error:
            logging.debug(
                f"Call to {self.model_name} with prompt: {prompt}\n"
                f"returned the following error:\n{val_error}"
            )
            return Failed(error=str(val_error))

    async def evaluate_answer_relevancy(
        self, eval_sample: EvaluationSample
    ) -> AnswerRelevancy | Failed:
        template = self.environment.get_template("answer_relevancy.txt")
        prompt = template.render(
            input=eval_sample.input,
            actual_output=eval_sample.actual_output,
            expected_output=eval_sample.expected_output,
        )
        return await self.call_llm(prompt, AnswerRelevancyPair)

    async def evaluate_completeness(
        self, eval_sample: EvaluationSample
    ) -> Completeness | Failed:
        template = self.environment.get_template("completeness.txt")
        prompt = template.render(
            input=eval_sample.input,
            actual_output=eval_sample.actual_output,
            expected_output=eval_sample.expected_output,
            contexts=eval_sample.references,
        )
        return await self.call_llm(prompt, CompletenessPair)

    async def evaluate_faithfulness(
        self, eval_sample: EvaluationSample
    ) -> Faithfulness | Failed:
        template = self.environment.get_template("faithfulness.txt")
        prompt = template.render(
            actual_output=eval_sample.actual_output,
            expected_output=eval_sample.expected_output,
            contexts=eval_sample.references,
        )
        return await self.call_llm(prompt, FaithfulnessPair)

    async def evaluate_usefulness(
        self, eval_sample: EvaluationSample
    ) -> Usefulness | Failed:
        template = self.environment.get_template("usefulness.txt")
        prompt = template.render(
            input=eval_sample.input,
            actual_output=eval_sample.actual_output,
            expected_output=eval_sample.expected_output,
        )
        return await self.call_llm(prompt, UsefulnessPair)

    async def evaluate_single_sample(
        self, eval_sample: EvaluationSample
    ) -> GroundedQAEvaluation:
        answer_relevancy = await self.evaluate_answer_relevancy(eval_sample)
        completeness = await self.evaluate_completeness(eval_sample)

        if isinstance(answer_relevancy, Failed):
            usefulness = Failed(error="answer_relevancy failed")
            faithfulness = Failed(error="answer_relevancy failed")
        else:
            if answer_relevancy.answer_relevancy is None:
                usefulness = await self.evaluate_usefulness(eval_sample)
                if isinstance(usefulness, Failed):
                    faithfulness = Failed(error="usefulness failed")
                elif usefulness.usefulness is None:
                    faithfulness = Faithfulness(
                        faithfulness_justification="", faithfulness=None
                    )
                else:
                    faithfulness = await self.evaluate_faithfulness(eval_sample)
            else:
                usefulness = Usefulness(usefulness_justification="", usefulness=None)
                faithfulness = await self.evaluate_faithfulness(eval_sample)

        positive_acceptance, negative_rejection = (
            get_positive_acceptance_negative_rejection(answer_relevancy, completeness)
        )

        return GroundedQAEvaluation(
            answer_relevancy=answer_relevancy,
            completeness=completeness,
            faithfulness=faithfulness,
            usefulness=usefulness,
            positive_acceptance=positive_acceptance,
            negative_rejection=negative_rejection,
        )

    async def async_evaluate_multiple_samples(
        self, eval_samples: List[EvaluationSample]
    ) -> List[GroundedQAEvaluation]:
        async with aiohttp.ClientSession() as _:
            evaluation_coroutines = [
                self.evaluate_single_sample(eval_sample) for eval_sample in eval_samples
            ]
            evaluations = await tqdm.gather(*evaluation_coroutines)
        return evaluations

    def evaluate_multiple_samples(
        self, eval_samples: List[EvaluationSample]
    ) -> List[GroundedQAEvaluation]:
        results = asyncio.run(self.async_evaluate_multiple_samples(eval_samples))
        self.logger.info(f"Cost: {self.cost:.4f}$")
        return results

    def evaluate(self, eval_samples: List[EvaluationSample]) -> EvaluationsAndReport:
        evaluations = self.evaluate_multiple_samples(eval_samples)
        ar_mean = np.mean(
            [
                e.answer_relevancy.answer_relevancy
                for e in evaluations
                if bool(e.answer_relevancy)
                and e.answer_relevancy.answer_relevancy is not None
            ]
        )
        ar_parsing_success = np.mean(
            [int(e.answer_relevancy is not None) for e in evaluations]
        )
        c_mean = np.mean(
            [
                e.completeness.completeness
                for e in evaluations
                if bool(e.completeness) and e.completeness.completeness is not None
            ]
        )
        c_parsing_success = np.mean(
            [int(e.completeness is not None) for e in evaluations]
        )
        f_mean = np.mean(
            [
                e.faithfulness.faithfulness
                for e in evaluations
                if bool(e.faithfulness) and e.faithfulness.faithfulness is not None
            ]
        )
        f_parsing_success = np.mean(
            [int(e.faithfulness is not None) for e in evaluations]
        )
        u_mean = np.mean(
            [
                e.usefulness.usefulness
                for e in evaluations
                if bool(e.usefulness) and e.usefulness.usefulness is not None
            ]
        )
        u_parsing_success = np.mean(
            [int(e.usefulness is not None) for e in evaluations]
        )
        pa_mean = np.mean(
            [
                e.positive_acceptance
                for e in evaluations
                if e.positive_acceptance is not None
            ]
        )
        nr_mean = np.mean(
            [
                e.negative_rejection
                for e in evaluations
                if e.negative_rejection is not None
            ]
        )
        mean = np.mean([ar_mean, c_mean, f_mean, u_mean, pa_mean, nr_mean])
        report = GroundedQAEvaluationReport(
            answer_relevancy=float(ar_mean),
            answer_relevancy_parsing_success=float(ar_parsing_success),
            completeness=float(c_mean),
            completeness_parsing_success=float(c_parsing_success),
            faithfulness=float(f_mean),
            faithfulness_parsing_success=float(f_parsing_success),
            usefulness=float(u_mean),
            usefulness_parse_success=float(u_parsing_success),
            positive_acceptance=float(pa_mean),
            negative_rejection=float(nr_mean),
            mean=mean,
        )
        return EvaluationsAndReport(evaluations=evaluations, report=report)



from typing import Optional
from unittest.mock import patch

import pytest

from grouse.dtos import (
    AnswerRelevancy,
    Completeness,
    EvaluationSample,
    ExpectedGroundedQAEvaluation,
)
from grouse.utils import get_positive_acceptance_negative_rejection, load_unit_tests


@pytest.mark.parametrize(
    "answer_relevancy, completeness, "
    "expected_positive_acceptance, expected_negative_rejection",
    [
        [
            AnswerRelevancy(
                answer_relevancy=None,
                answer_relevancy_justification="",
                answer_affirms_no_document_answers=False,
            ),
            Completeness(completeness=None, completeness_justification=""),
            1,
            1,
        ],
        [
            AnswerRelevancy(
                answer_relevancy=None,
                answer_relevancy_justification="",
                answer_affirms_no_document_answers=False,
            ),
            Completeness(completeness=5, completeness_justification=""),
            0,
            None,
        ],
        [
            AnswerRelevancy(
                answer_relevancy=5,
                answer_relevancy_justification="",
                answer_affirms_no_document_answers=False,
            ),
            Completeness(completeness=None, completeness_justification=""),
            None,
            0,
        ],
        [
            AnswerRelevancy(
                answer_relevancy=5,
                answer_relevancy_justification="",
                answer_affirms_no_document_answers=False,
            ),
            Completeness(completeness=5, completeness_justification=""),
            None,
            None,
        ],
    ],
)
def test_get_positive_acceptance_negative_rejection(
    answer_relevancy: Optional[int],
    completeness: Optional[int],
    expected_positive_acceptance: Optional[int],
    expected_negative_rejection: Optional[int],
) -> None:
    pa, nr = get_positive_acceptance_negative_rejection(answer_relevancy, completeness)
    assert pa == expected_positive_acceptance
    assert nr == expected_negative_rejection


def test_load_unit_tests() -> None:
    with patch(
        "grouse.utils.load_dataset",
        return_value={
            "train": [
                {
                    "input": "Quel est la capitale de la France ?",
                    "actual_output": "Paris[1]",
                    "expected_output": "Paris[1]",
                    "references": ["Paris"],
                    "metadata": {},
                    "conditions": {
                        "answer_relevancy_condition": "==5",
                        "completeness_condition": "==5",
                        "faithfulness_condition": "==1",
                        "usefulness_condition": "==None",
                    },
                }
            ],
            "test": [
                {
                    "input": "Quel est la capitale de la France ?",
                    "actual_output": "Paris[1]",
                    "expected_output": "Paris[1]",
                    "references": ["Paris"],
                    "metadata": {},
                    "conditions": {
                        "answer_relevancy_condition": "==5",
                        "completeness_condition": "==5",
                        "faithfulness_condition": "==1",
                        "usefulness_condition": "==None",
                    },
                }
            ],
        },
    ):
        samples, conditions = load_unit_tests()
        assert isinstance(samples, list)
        assert isinstance(conditions, list)
        assert isinstance(samples[0], EvaluationSample)
        assert isinstance(conditions[0], ExpectedGroundedQAEvaluation)



from typing import List, Literal, Optional, Union

from pydantic import BaseModel, Field
from typing_extensions import override


class Failed(BaseModel):
    """
    A sentinel class used to distinguish failed request results from results
    with the value None (which may have different behavior).
    """

    error: Optional[str] = None

    def __bool__(self) -> Literal[False]:
        return False

    def __int__(self) -> int:
        return 0

    @override
    def __repr__(self) -> str:
        return "FAILED"


# These models are used by instructor to enforce the output json schema
# Be careful if you change these models, as the field names are optimized for the prompt
class Score(BaseModel):
    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def __str__(self):
        return self.__repr__()


class AnswerRelevancy(Score):
    answer_affirms_no_document_answers: bool
    answer_relevancy_justification: str
    answer_relevancy: Optional[int] = Field(
        None,
        description="Relevancy score of the answer from 1 to 5 or None",
    )


class Completeness(Score):
    completeness_justification: str
    completeness: Optional[int] = Field(
        None,
        description="Completeness score of the answer from 1 to 5 or None",
    )


class Faithfulness(Score):
    faithfulness_justification: str
    faithfulness: Optional[int] = Field(
        None,
        description="Faithfulness score of the answer in 0, 1 or None",
    )


class Usefulness(Score):
    usefulness_justification: str
    usefulness: Optional[int] = Field(
        None, description="Usefulness score of the answer in 0, 1 or None"
    )


# These models are used by instructor to enforce the input json schema
# Don't try to use generic types here, as they are not supported by instuctor
# because the name of the model would have '[' and ']' characters
class AnswerRelevancyPair(BaseModel):
    answer_1: AnswerRelevancy
    answer_2: AnswerRelevancy


class CompletenessPair(BaseModel):
    answer_1: Completeness
    answer_2: Completeness


class FaithfulnessPair(BaseModel):
    answer_1: Faithfulness
    answer_2: Faithfulness


class UsefulnessPair(BaseModel):
    answer_1: Usefulness
    answer_2: Usefulness


ScorePair = Union[
    AnswerRelevancyPair, CompletenessPair, FaithfulnessPair, UsefulnessPair
]


# Evaluation DTOs
class EvaluationSample(BaseModel):
    """Model representing the input, the output generated by the model we are
    evaluating, the expected output and the references."""

    input: str
    actual_output: str
    expected_output: str
    references: List[str]
    metadata: Optional[dict] = None


class GroundedQAEvaluationReport(BaseModel):
    """Model that contains the average scores of the evaluations

    Args:
        answer_relevancy (float): Average relevancy score.
        answer_relevancy_parsing_success (float): Success rate of parsing relevancy
        JSONs.
        completeness (float): Average completeness score.
        completeness_parsing_success (float): Success rate of parsing completeness
        JSONs.
        faithfulness (float): Average faithfulness score.
        faithfulness_parsing_success (float): Success rate of parsing faithfulness
        JSONs.
        usefulness (float): Average usefulness score.
        usefulness_parse_success (float): Success rate of parsing usefulness JSONs.
        positive_acceptance (float): Positive acceptance rate.
        negative_rejection (float): Negative rejection rate.
        mean (float): Average of answer_relevancy, completeness, faithfulness,
        usefulness, positive_acceptance and negative_rejection.
    """

    answer_relevancy: float
    answer_relevancy_parsing_success: float
    completeness: float
    completeness_parsing_success: float
    faithfulness: float
    faithfulness_parsing_success: float
    usefulness: float
    usefulness_parse_success: float
    positive_acceptance: float
    negative_rejection: float
    mean: float


class GroundedQAEvaluation(BaseModel):
    """Model representing the evaluation results for grounded QA on one sample."""

    answer_relevancy: AnswerRelevancy | Failed
    completeness: Completeness | Failed
    faithfulness: Faithfulness | Failed
    usefulness: Usefulness | Failed
    positive_acceptance: Optional[int] | Failed
    negative_rejection: Optional[int] | Failed


class EvaluationsAndReport(BaseModel):
    """
    Final output of the evaluation containing the individual evaluations and
    the aggregated results.
    """

    evaluations: List[GroundedQAEvaluation]
    report: GroundedQAEvaluationReport


# Meta Evaluation DTOs
class ExpectedGroundedQAEvaluation(BaseModel):
    """Model used to define the conditions that need to be verified by
    the LLM evaluator output that we are meta-evaluating.

    Args:
        answer_relevancy_condition (str): Condition in the format "<operator><value>".
        For example, "==5", ">=3" or "==None".
        completeness_condition: Same as above.
        faithfulness_condition: Same as above.
        usefulness_condition: Same as above.
    """

    answer_relevancy_condition: str
    completeness_condition: str
    faithfulness_condition: str
    usefulness_condition: str


class MetaTestCase(BaseModel):
    """DTO formatting an individual test case for the meta-evaluator."""

    evaluation_sample: EvaluationSample
    actual_evaluation: GroundedQAEvaluation
    expected_evaluation: ExpectedGroundedQAEvaluation


class MetaTestCaseResult(BaseModel):
    """Individual test case result for the meta-evaluator."""

    answer_relevancy: bool | Failed
    completeness: bool | Failed
    faithfulness: bool | Failed
    usefulness: bool | Failed
    positive_acceptance: bool | Failed
    negative_rejection: bool | Failed


class MetaEvalReport(BaseModel):
    """Aggregated Summary of the meta-evaluation results.

    Args:
        answer_relevancy_success (float): Success rate of answer relevancy conditions.
        completeness_success (float): Success rate of completeness conditions.
        faithfulness_success (float): Success rate of faithfulness conditions.
        usefulness_success (float): Success rate of usefulness conditions.
        positive_acceptance_success (float): Success rate of
        positive acceptance conditions.
        negative_rejection_success (float): Success rate of
        negative rejection conditions.
    """

    answer_relevancy_success: float
    completeness_success: float
    faithfulness_success: float
    usefulness_success: float
    positive_acceptance_success: float
    negative_rejection_success: float
    total: float


class MetaEvaluationsAndReport(BaseModel):
    """Final output of the meta-evaluation containing the individual evaluations and
    the aggregated results."""

    evaluations: List[MetaTestCaseResult]
    report: MetaEvalReport



from typing import List, Optional

from grouse.dtos import (
    Failed,
    MetaEvalReport,
    MetaEvaluationsAndReport,
    MetaTestCase,
    MetaTestCaseResult,
    Score,
)
from grouse.utils import get_positive_acceptance_negative_rejection


class MetaEvaluator:
    def __init__(self) -> None:
        pass

    @staticmethod
    def compare(value: Optional[float], condition: str) -> bool:
        if value is None:
            return condition == "==None"
        else:
            if condition.endswith("None"):
                return False
            elif condition.startswith(">="):
                return value >= float(condition[2:])
            elif condition.startswith("<="):
                return value <= float(condition[2:])
            elif condition.startswith(">"):
                return value > float(condition[1:])
            elif condition.startswith("<"):
                return value < float(condition[1:])
            elif condition[:2] == "==":
                return value == float(condition[2:])
            else:
                raise ValueError("Invalid condition")

    def __get_result(self, score: Score, score_name: str, condition: str) -> bool:
        if isinstance(score, Failed):
            return Failed(error=score.error)
        return self.compare(getattr(score, score_name), condition)

    def evaluate_single_test_case(self, test_case: MetaTestCase) -> MetaTestCaseResult:
        answer_relevancy_result = self.__get_result(
            test_case.actual_evaluation.answer_relevancy,
            "answer_relevancy",
            test_case.expected_evaluation.answer_relevancy_condition,
        )
        completeness_result = self.__get_result(
            test_case.actual_evaluation.completeness,
            "completeness",
            test_case.expected_evaluation.completeness_condition,
        )
        faithfulness_result = self.__get_result(
            test_case.actual_evaluation.faithfulness,
            "faithfulness",
            test_case.expected_evaluation.faithfulness_condition,
        )
        usefulness_result = self.__get_result(
            test_case.actual_evaluation.usefulness,
            "usefulness",
            test_case.expected_evaluation.usefulness_condition,
        )

        if isinstance(
            test_case.actual_evaluation.answer_relevancy, Failed
        ) or isinstance(test_case.actual_evaluation.completeness, Failed):
            positive_acceptance_result = Failed()
            negative_rejection_result = Failed()
        else:
            if test_case.expected_evaluation.answer_relevancy_condition == "==None":
                if test_case.expected_evaluation.completeness_condition == "==None":
                    positive_acceptance_condition = "==1"
                    negative_rejection_condition = "==1"
                else:
                    positive_acceptance_condition = "==0"
                    negative_rejection_condition = "==None"
            else:
                if test_case.expected_evaluation.completeness_condition == "==None":
                    positive_acceptance_condition = "==None"
                    negative_rejection_condition = "==0"
                else:
                    positive_acceptance_condition = "==None"
                    negative_rejection_condition = "==None"

            evaluated_positive_acceptance, evaluated_negative_rejection = (
                get_positive_acceptance_negative_rejection(
                    test_case.actual_evaluation.answer_relevancy,
                    test_case.actual_evaluation.completeness,
                )
            )
            positive_acceptance_result = self.compare(
                evaluated_positive_acceptance, positive_acceptance_condition
            )
            negative_rejection_result = self.compare(
                evaluated_negative_rejection, negative_rejection_condition
            )

        return MetaTestCaseResult(
            answer_relevancy=answer_relevancy_result,
            completeness=completeness_result,
            faithfulness=faithfulness_result,
            usefulness=usefulness_result,
            positive_acceptance=positive_acceptance_result,
            negative_rejection=negative_rejection_result,
        )

    def evaluate_multiple_test_cases(
        self, test_cases: list[MetaTestCase]
    ) -> List[MetaTestCaseResult]:
        return [self.evaluate_single_test_case(test_case) for test_case in test_cases]

    def evaluate(self, test_cases: List[MetaTestCase]) -> MetaEvaluationsAndReport:
        meta_evaluations = self.evaluate_multiple_test_cases(test_cases)
        answer_relevancy_success = float(
            sum([int(e.answer_relevancy) for e in meta_evaluations])
            / len(meta_evaluations)
        )
        completeness_success = float(
            sum([int(e.completeness) for e in meta_evaluations]) / len(meta_evaluations)
        )
        faithfulness_success = float(
            sum([int(e.faithfulness) for e in meta_evaluations]) / len(meta_evaluations)
        )
        usefulness_success = float(
            sum([int(e.usefulness) for e in meta_evaluations]) / len(meta_evaluations)
        )
        positive_acceptance_success = float(
            sum([int(e.positive_acceptance) for e in meta_evaluations])
            / len(meta_evaluations)
        )
        negative_rejection_success = float(
            sum([int(e.negative_rejection) for e in meta_evaluations])
            / len(meta_evaluations)
        )
        total = (
            answer_relevancy_success
            + completeness_success
            + faithfulness_success
            + usefulness_success
            + positive_acceptance_success
            + negative_rejection_success
        ) / 6
        return MetaEvaluationsAndReport(
            evaluations=meta_evaluations,
            report=MetaEvalReport(
                answer_relevancy_success=answer_relevancy_success,
                completeness_success=completeness_success,
                faithfulness_success=faithfulness_success,
                usefulness_success=usefulness_success,
                positive_acceptance_success=positive_acceptance_success,
                negative_rejection_success=negative_rejection_success,
                total=total,
            ),
        )



import json
import os
from typing import Optional

import click
import jsonlines

from grouse.dtos import EvaluationSample, MetaTestCase, MetaTestCaseResult
from grouse.grounded_qa_evaluator import GroundedQAEvaluator
from grouse.meta_evaluator import MetaEvaluator
from grouse.plot import plot_matrices
from grouse.register_models import register_models
from grouse.utils import NanConverter, load_unit_tests

register_models()


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.argument("dataset_path", type=str)
@click.argument("output_dir_path", type=str)
@click.option(
    "--evaluator_model_name",
    type=str,
    help=(
        "Name of the evaluator model. It can be any LiteLLM model. "
        "The default model is GPT-4."
    ),
    default="gpt-4",
)
@click.option(
    "--prompts_path",
    type=str,
    help=(
        "Path to the folder containing the prompts of the evaluator. "
        "By default, the prompts are those optimized for GPT-4."
    ),
    default=None,
)
def evaluate(
    dataset_path: str,
    output_dir_path: str,
    evaluator_model_name: Optional[str] = None,
    prompts_path: Optional[str] = None,
) -> None:
    """Evaluate models on grounded question answering using any LiteLLM model.
    The default model is GPT-4.

    Args:
        DATASET_PATH (str): Path to jsonlines file with references, input,
        actual_output (generation from the model to evaluate) and expected_output.
        OUTPUT_DIR_PATH (str): Path to directory where results report and
        evaluations are saved.
    """
    evaluator = GroundedQAEvaluator(
        model_name=evaluator_model_name, prompts_path=prompts_path
    )
    eval_samples = []
    with jsonlines.open(dataset_path) as reader:
        for obj in reader:
            eval_samples.append(EvaluationSample(**obj))

    results = evaluator.evaluate(eval_samples)

    os.makedirs(output_dir_path, exist_ok=True)
    with open(
        os.path.join(output_dir_path, "report.json"), "w", encoding="utf-8"
    ) as file:
        json.dump(results.report.model_dump(mode="json"), file, cls=NanConverter)

    with jsonlines.open(
        os.path.join(output_dir_path, "evaluations.jsonl"), "w"
    ) as writer:
        for evaluation in results.evaluations:
            writer.write(evaluation.model_dump(mode="json"))


@cli.command()
@click.argument("model_name", type=str)
@click.argument("output_dir_path", type=str)
@click.option(
    "--prompts_path",
    type=str,
    help=(
        "Path to the folder containing the prompts of the evaluator. "
        "By default, the prompts are those optimized for GPT-4."
    ),
    default=None,
)
def meta_evaluate(
    model_name: str, output_dir_path: str, prompts_path: Optional[str] = None
) -> None:
    """Evaluate evaluators on GroUSE unit tests.

    Args:
        MODEL_NAME (str): Name of model available through LiteLLM.
        OUTPUT_DIR_PATH (str): Path to directory where results report and
        unit test results are saved.
    """
    evaluation_samples, conditions = load_unit_tests()

    evaluator = GroundedQAEvaluator(model_name, prompts_path=prompts_path)
    evaluations = evaluator.evaluate_multiple_samples(evaluation_samples)

    meta_evaluator = MetaEvaluator()

    meta_test_cases = []
    for sample, evaluation, condition in zip(
        evaluation_samples, evaluations, conditions
    ):
        meta_test_cases.append(
            MetaTestCase(
                evaluation_sample=sample,
                actual_evaluation=evaluation,
                expected_evaluation=condition,
            )
        )

    meta_evaluations = meta_evaluator.evaluate(meta_test_cases)

    os.makedirs(output_dir_path, exist_ok=True)
    with open(
        os.path.join(output_dir_path, "report.json"), "w", encoding="utf-8"
    ) as file:
        json.dump(
            meta_evaluations.report.model_dump(mode="json"), file, cls=NanConverter
        )

    with jsonlines.open(
        os.path.join(output_dir_path, "meta_evaluations.jsonl"), "w"
    ) as writer:
        for evaluation in meta_evaluations.evaluations:
            writer.write(evaluation.model_dump(mode="json"))


@cli.command()
@click.argument("meta_test_results_path", type=str)
def plot(meta_test_results_path: str) -> None:
    """Create matrix plots for the four main metrics

    Args:
        META_TEST_RESULTS_PATH (str): Path to meta evaluation results in
        jsonlines format.
    """
    evaluation_samples, _ = load_unit_tests()

    results = []
    with jsonlines.open(meta_test_results_path, "r") as reader:
        for obj in reader:
            results.append(MetaTestCaseResult(**obj))
    plot_matrices(evaluation_samples, results)



import math
from json import JSONEncoder
from typing import Any, List, Optional, Tuple

from datasets import load_dataset

from grouse.dtos import (
    AnswerRelevancy,
    Completeness,
    EvaluationSample,
    ExpectedGroundedQAEvaluation,
    Failed,
)

DATASET_NAME = "illuin/grouse"


def nan_to_none(obj: Any) -> Any:
    if isinstance(obj, dict):
        return {k: nan_to_none(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [nan_to_none(v) for v in obj]
    elif isinstance(obj, float) and math.isnan(obj):
        return None
    return obj


class NanConverter(JSONEncoder):
    def encode(self, obj: Any, *args, **kwargs) -> Any:
        return super().encode(nan_to_none(obj), *args, **kwargs)

    def iterencode(self, obj: Any, *args, **kwargs) -> Any:
        return super().iterencode(nan_to_none(obj), *args, **kwargs)


def get_positive_acceptance_negative_rejection(
    answer_relevancy: AnswerRelevancy | Failed,
    completeness: Completeness | Failed,
) -> Tuple[Optional[int] | Failed, Optional[int] | Failed]:
    if isinstance(answer_relevancy, Failed):
        return Failed(error="answer relevancy failed"), Failed(
            error="answer relevancy failed"
        )
    elif isinstance(completeness, Failed):
        return Failed(error="completeness failed"), Failed(error="completeness failed")
    else:
        if answer_relevancy.answer_relevancy is None:
            if completeness.completeness is None:
                positive_acceptance = 1
                negative_rejection = 1
            else:
                positive_acceptance = 0
                negative_rejection = None
        else:
            if completeness.completeness is None:
                positive_acceptance = None
                negative_rejection = 0
            else:
                positive_acceptance = None
                negative_rejection = None
        return positive_acceptance, negative_rejection


def load_unit_tests() -> (
    Tuple[List[EvaluationSample], List[ExpectedGroundedQAEvaluation]]
):
    unit_tests = load_dataset(DATASET_NAME)["test"]
    evaluation_samples = []
    conditions = []

    for unit_test in unit_tests:
        evaluation_samples.append(
            EvaluationSample(
                input=unit_test["input"],
                actual_output=unit_test["actual_output"],
                expected_output=unit_test["expected_output"],
                references=unit_test["references"],
                metadata=unit_test["metadata"],
            )
        )
        conditions.append(ExpectedGroundedQAEvaluation(**unit_test["conditions"]))

    return evaluation_samples, conditions



from typing import List

import matplotlib.colors as mcolors
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.legend_handler import HandlerBase

from grouse.dtos import (
    EvaluationSample,
    Failed,
    MetaTestCaseResult,
)

VALUE_COLORS = {1: "tab:blue", 0: "tab:red", 2: "tab:orange"}
VALUE_LABELS = {1: "Test passed", 0: "Test failed", 2: "Output wrong format"}
CMAP = plt.cm.colors.ListedColormap([VALUE_COLORS[0], VALUE_COLORS[1], VALUE_COLORS[2]])
NORM = mcolors.BoundaryNorm(boundaries=[0, 1, 2, 3], ncolors=3)


def add_custom_legend(
    hatch_legend: bool = True,
    value_colors=VALUE_COLORS,
    value_labels=VALUE_LABELS,
    fig=None,
    ax=None,
    *args,
    **kwargs,
):
    if fig is None and ax is None:
        raise ValueError("Both fig and ax can't be None")
    if fig is not None and ax is not None:
        raise ValueError("Both fig and ax can't be different than None")
    fig_or_ax = fig or ax

    patch_size = 20
    list_patches = []
    for key in value_colors.keys():
        list_patches.append(
            patches.Rectangle(
                (0, 0),
                patch_size,
                patch_size,
                color=value_colors[key],
                label=value_labels[key],
            )
        )
    if hatch_legend:
        list_patches.append(
            patches.Rectangle(
                (0, 0),
                patch_size,
                patch_size,
                facecolor="white",
                label="Theoretically won't happen",
                hatch="/////",
            )
        )
    fig_or_ax.legend(
        *args, handles=list_patches, handlelength=1, handleheight=1, **kwargs
    )


def plot_matrix(
    values, row_names, title, ax, hatch_rows, show_yticks=False, show_xlabel=False
):
    matrix = []
    for row_index in range(16):
        column = []
        for column_index in range(9):
            value = values[column_index * 16 + row_index]
            if np.isnan(value):
                column.append(2)
            else:
                column.append(value)
        matrix.append(column)
    matrix = np.array(matrix)

    # Adds white lines between squares
    for k in range(matrix.shape[0] - 1):
        ax.axhline(k + 0.5, color="white", linewidth=2)
    for k in range(matrix.shape[1] - 1):
        ax.axvline(k + 0.5, color="white", linewidth=2)

    ax.imshow(matrix, cmap=CMAP, norm=NORM, interpolation="nearest")
    for row in hatch_rows:
        for column in range(matrix.shape[1]):
            ax.add_patch(
                patches.Rectangle(
                    (column - 0.5, row - 0.5),
                    1,
                    1,
                    hatch="///",
                    fill=False,
                    snap=False,
                    linewidth=0,
                )
            )

    # display all ticks on the x-axis
    ax.set_xticks(np.arange(matrix.shape[1]))
    ax.set_xticklabels(np.arange(matrix.shape[1]))

    # display all ticks on the y-axis
    ax.set_yticks(np.arange(matrix.shape[0]))
    ax.set_yticklabels(np.arange(matrix.shape[0]))

    xticklabels = ax.get_xticklabels()
    xtick_positions = ax.get_xticks()

    ax.set_xticklabels(
        list(range(len(xticklabels))), rotation=0, ha="center", fontsize=10
    )

    # Plot a circle around each xtick label
    for pos in xtick_positions:
        circle = plt.Circle((pos, 16.3), 0.4, color="black", fill=False, clip_on=False)
        ax.add_artist(circle)

    # Change the xticks and yticks labels
    ax.set_xticklabels(list(range(1, matrix.shape[1] + 1)), fontsize=14)

    if show_xlabel:
        ax.set_xlabel("Test samples", labelpad=10, fontsize=16)

    if show_yticks:
        ax.set_yticklabels(row_names, fontsize=14)
        ax.set_ylabel("Test type", labelpad=10, fontsize=16)
    else:
        ax.set_yticklabels(list(range(1, matrix.shape[0] + 1)), fontsize=14)
    ax.set_title(title, fontsize=18)


def build_circle_legend(letter_to_label_dict):
    handler_map = {}
    fake_objects = []
    labels = []
    for letter, label in letter_to_label_dict.items():
        HandlerBase  # to avoid having the import removed by isort or something

        exec(
            f"""class myclass_{letter}:
        pass"""
        )
        exec(f"fake_objects.append(myclass_{letter}())")

        exec(
            f"""class LetterLegendHandler{letter}(HandlerBase):
    def create_artists(self, legend, orig_handle, xdescent, ydescent, width, height, fontsize, trans):
        text = "{letter}"
        text_artist = plt.Text(
            xdescent + 0.5 * width, ydescent + 0.35 * height, text, fontsize=fontsize, ha="center", va="center"
        )
        circle_artist = plt.Circle((xdescent + 0.5 * width, ydescent + 0.4 * height), width * 0.32, fill=False)
        return [text_artist, circle_artist]"""
        )

        exec(f"handler_map[myclass_{letter}] =  LetterLegendHandler{letter}()")
        labels.append(label)

    return fake_objects, labels, handler_map


def process_value(value):
    if isinstance(value, Failed):
        return 2
    else:
        return int(value)


def plot_matrices(
    evaluation_samples: List[EvaluationSample],
    meta_evaluations: List[MetaTestCaseResult],
) -> None:
    questions_seen = set()
    unique_questions = []
    test_types_seen = set()
    unique_test_types = []

    for evaluation_sample in evaluation_samples:
        if evaluation_sample.input not in questions_seen:
            unique_questions.append(evaluation_sample.input)
            questions_seen.add(evaluation_sample.input)
        test_type = evaluation_sample.metadata.get("test_type", "")
        if test_type not in test_types_seen:
            unique_test_types.append(test_type)
            test_types_seen.add(test_type)

    fig, axes = plt.subplots(nrows=1, ncols=4, figsize=(40, 10))
    fig.subplots_adjust(bottom=0.5, left=0.2)

    answer_relevancy_values = [
        process_value(e.answer_relevancy) for e in meta_evaluations
    ]
    completeness_values = [process_value(e.completeness) for e in meta_evaluations]
    faithfulness_values = [process_value(e.faithfulness) for e in meta_evaluations]
    usefulness_values = [process_value(e.usefulness) for e in meta_evaluations]

    plot_matrix(
        answer_relevancy_values,
        unique_test_types,
        "Answer Relevancy",
        axes[0],
        [],
        show_yticks=True,
        show_xlabel=True,
    )
    plot_matrix(
        completeness_values,
        unique_test_types,
        "Completeness",
        axes[1],
        [],
        show_yticks=False,
        show_xlabel=True,
    )
    plot_matrix(
        usefulness_values,
        unique_test_types,
        "Usefulness",
        axes[2],
        [0, 3, 5, 7, 8, 9, 13, 14, 15],
        show_yticks=False,
        show_xlabel=True,
    )
    plot_matrix(
        faithfulness_values,
        unique_test_types,
        "Faithfulness",
        axes[3],
        [1, 4, 10],
        show_yticks=False,
        show_xlabel=True,
    )

    bbox_to_anchor = (0, -0.7)
    custom_legend_loc = "lower left"
    value_labels = {0: "Test failed", 1: "Test passed", 2: "Output Wrong Format"}
    add_custom_legend(
        hatch_legend=True,
        value_colors=VALUE_COLORS,
        value_labels=value_labels,
        ax=axes[2],
        bbox_to_anchor=bbox_to_anchor,
        loc=custom_legend_loc,
        fontsize=18,
    )
    int_to_label_dict = {
        str(i + 1): question for (i, question) in enumerate(unique_questions)
    }
    test_questions_legend_bbox_to_anchor = (0.3, 0.3)
    fake_objects_circle, labels_circle, handler_map_circle = build_circle_legend(
        int_to_label_dict
    )
    fig.legend(
        handles=fake_objects_circle,
        labels=labels_circle,
        handler_map=handler_map_circle,
        bbox_to_anchor=test_questions_legend_bbox_to_anchor,
        loc="center",
        title="Test questions",
        title_fontproperties={"weight": "bold"},
    )
    # return fig, axes
    plt.show()



import asyncio
from unittest.mock import patch

from grouse import EvaluationSample, GroundedQAEvaluator
from grouse.dtos import (
    AnswerRelevancy,
    Completeness,
    EvaluationsAndReport,
    Failed,
    Faithfulness,
    GroundedQAEvaluation,
    GroundedQAEvaluationReport,
    Usefulness,
)

TEST_MODEL = "gpt-4o-mini"


class TestGroundedQAEvaluator:
    def setup_method(self) -> None:
        self.evaluator = GroundedQAEvaluator(model_name=TEST_MODEL)

    def test_evaluate_answer_relevancy(self) -> None:
        with patch.object(
            GroundedQAEvaluator,
            "call_llm",
            return_value=AnswerRelevancy(
                answer_relevancy=5,
                answer_affirms_no_document_answers=False,
                answer_relevancy_justification="Justification",
            ),
        ):
            ar = asyncio.run(
                self.evaluator.evaluate_answer_relevancy(
                    eval_sample=EvaluationSample(
                        input="Quel est la capitale de la France ?",
                        actual_output="Paris",
                        expected_output="Paris",
                        references=["Paris"],
                    )
                )
            )
            assert isinstance(ar, AnswerRelevancy)
            assert isinstance(ar.answer_affirms_no_document_answers, bool)
            assert isinstance(ar.answer_relevancy_justification, str)
            assert isinstance(ar.answer_relevancy, int)
            assert ar.answer_relevancy == 5

    def test_evaluate_completeness(self) -> None:
        with patch.object(
            GroundedQAEvaluator,
            "call_llm",
            return_value=Completeness(
                completeness=5,
                completeness_justification="Justification",
            ),
        ):
            comp = asyncio.run(
                self.evaluator.evaluate_completeness(
                    eval_sample=EvaluationSample(
                        input="Quel est la capitale de la France ?",
                        actual_output="Paris",
                        expected_output="Paris",
                        references=["Paris"],
                    )
                )
            )
            assert isinstance(comp, Completeness)
            assert isinstance(comp.completeness_justification, str)
            assert isinstance(comp.completeness, int)
            assert comp.completeness == 5

    def test_evaluate_faithfulness(self) -> None:
        with patch.object(
            GroundedQAEvaluator,
            "call_llm",
            return_value=Faithfulness(
                faithfulness=1,
                faithfulness_justification="Justification",
            ),
        ):
            faith = asyncio.run(
                self.evaluator.evaluate_faithfulness(
                    eval_sample=EvaluationSample(
                        input="Quel est la capitale de la France ?",
                        actual_output="Paris[1]",
                        expected_output="Paris[1]",
                        references=["Paris"],
                    )
                )
            )
            assert isinstance(faith, Faithfulness)
            assert isinstance(faith.faithfulness_justification, str)
            assert isinstance(faith.faithfulness, int)
            assert faith.faithfulness == 1

    def test_evaluate_usefulness(self) -> None:
        with patch.object(
            GroundedQAEvaluator,
            "call_llm",
            return_value=Usefulness(
                usefulness=None,
                usefulness_justification="Justification",
            ),
        ):
            usef = asyncio.run(
                self.evaluator.evaluate_usefulness(
                    eval_sample=EvaluationSample(
                        input="Quel est la capitale de la France ?",
                        actual_output="Paris[1]",
                        expected_output="Paris[1]",
                        references=["Paris"],
                    )
                )
            )
            assert isinstance(usef, Usefulness)
            assert isinstance(usef.usefulness_justification, str)
            assert usef.usefulness is None

    def test_evaluate_single_sample(self) -> None:
        with (
            patch.object(
                GroundedQAEvaluator,
                "evaluate_answer_relevancy",
                return_value=AnswerRelevancy(
                    answer_relevancy=5,
                    answer_affirms_no_document_answers=False,
                    answer_relevancy_justification="Justification",
                ),
            ),
            patch.object(
                GroundedQAEvaluator,
                "evaluate_completeness",
                return_value=Completeness(
                    completeness=5,
                    completeness_justification="Justification",
                ),
            ),
            patch.object(
                GroundedQAEvaluator,
                "evaluate_faithfulness",
                return_value=Faithfulness(
                    faithfulness=1,
                    faithfulness_justification="Justification",
                ),
            ),
            patch.object(
                GroundedQAEvaluator,
                "evaluate_usefulness",
                return_value=Usefulness(
                    usefulness=None,
                    usefulness_justification="Justification",
                ),
            ),
        ):
            evaluation = asyncio.run(
                self.evaluator.evaluate_single_sample(
                    eval_sample=EvaluationSample(
                        input="Quel est la capitale de la France ?",
                        actual_output="Paris[1]",
                        expected_output="Paris[1]",
                        references=["Paris"],
                    )
                )
            )
            assert isinstance(evaluation, GroundedQAEvaluation)

    def test_evaluate_single_sample_with_failed_parsing(self) -> None:
        with patch.object(
            GroundedQAEvaluator,
            "call_llm",
            return_value=Failed(),
        ):
            evaluation = asyncio.run(
                self.evaluator.evaluate_single_sample(
                    eval_sample=EvaluationSample(
                        input="Quel est la capitale de la France ?",
                        actual_output="Paris[1]",
                        expected_output="Paris[1]",
                        references=["Paris"],
                    )
                )
            )
            assert isinstance(evaluation, GroundedQAEvaluation)
            assert isinstance(evaluation.answer_relevancy, Failed)
            assert isinstance(evaluation.completeness, Failed)
            assert isinstance(evaluation.faithfulness, Failed)
            assert isinstance(evaluation.usefulness, Failed)
            assert isinstance(evaluation.positive_acceptance, Failed)
            assert isinstance(evaluation.negative_rejection, Failed)

    def test_evaluate(self) -> None:
        with patch.object(
            GroundedQAEvaluator,
            "evaluate_single_sample",
            return_value=GroundedQAEvaluation(
                answer_relevancy=AnswerRelevancy(
                    answer_relevancy=5,
                    answer_affirms_no_document_answers=False,
                    answer_relevancy_justification="Justification",
                ),
                completeness=Completeness(
                    completeness=5,
                    completeness_justification="Justification",
                ),
                faithfulness=Faithfulness(
                    faithfulness=1,
                    faithfulness_justification="Justification",
                ),
                usefulness=Usefulness(
                    usefulness=None,
                    usefulness_justification="Justification",
                ),
                positive_acceptance=None,
                negative_rejection=None,
            ),
        ):
            eval_samples = [
                EvaluationSample(
                    input="Quel est la capitale de la France ?",
                    actual_output="Paris[1]",
                    expected_output="Paris[1]",
                    references=["Paris"],
                )
            ]
            evaluations_and_report = self.evaluator.evaluate(eval_samples)
            assert isinstance(evaluations_and_report, EvaluationsAndReport)
            evaluations, report = (
                evaluations_and_report.evaluations,
                evaluations_and_report.report,
            )
            assert isinstance(evaluations, list)
            assert len(evaluations) == 1
            assert isinstance(evaluations[0], GroundedQAEvaluation)
            assert isinstance(report, GroundedQAEvaluationReport)
